"""Provides high-level DNDarray initialization functions"""

import numpy as np
import torch
import warnings

from typing import Callable, Iterable, Optional, Sequence, Tuple, Type, Union, List

from .communication import MPI, sanitize_comm, Communication
from .devices import Device
from .dndarray import DNDarray
from .memory import sanitize_memory_layout
from .sanitation import sanitize_in, sanitize_sequence
from .stride_tricks import sanitize_axis, sanitize_shape
from .types import datatype

from . import devices
from . import types

__all__ = [
    "arange",
    "array",
    "asarray",
    "empty",
    "empty_like",
    "eye",
    "from_partitioned",
    "from_partition_dict",
    "full",
    "full_like",
    "linspace",
    "logspace",
    "meshgrid",
    "ones",
    "ones_like",
    "zeros",
    "zeros_like",
]


def arange(
    *args: Union[int, float],
    dtype: Optional[Type[datatype]] = None,
    split: Optional[int] = None,
    device: Optional[Union[str, Device]] = None,
    comm: Optional[Communication] = None,
) -> DNDarray:
    """
    Return evenly spaced values within a given interval.

    Values are generated within the half-open interval ``[start, stop)`` (in other words, the interval including `start`
    but excluding `stop`). For integer arguments the function is equivalent to the Python built-in `range
    <http://docs.python.org/lib/built-in-funcs.html>`_ function, but returns a array rather than a list.
    When using a non-integer step, such as 0.1, the results may be inconsistent due to being subject to numerical
    rounding. In the cases the usage of :func:`linspace` is recommended.
    For floating point arguments, the length of the result is :math:`\\lceil(stop-start)/step\\rceil`.
    Again, due to floating point rounding, this rule may result in the last element of `out` being greater than `stop`
    by machine epsilon.

    Parameters
    ----------
    start : scalar, optional
        Start of interval.  The interval includes this value.  The default start value is 0.
    stop : scalar
        End of interval.  The interval does not include this value, except in some cases where ``step`` is not an
        integer and floating point round-off affects the length of ``out``.
    step : scalar, optional
        Spacing between values.  For any output ``out``, this is the distance between two adjacent values,
        ``out[i+1]-out[i]``. The default step size is 1. If ``step`` is specified as a position argument, ``start``
        must also be given.
    dtype : datatype, optional
        The type of the output array.  If `dtype` is not given, it is automatically inferred from the other input
        arguments.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str, optional
        Specifies the device the array shall be allocated on, defaults to globally set default device.
    comm : Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.

    See Also
    --------
    :func:`linspace` : Evenly spaced numbers with careful handling of endpoints.

    Examples
    --------
    >>> ht.arange(3)
    DNDarray([0, 1, 2], dtype=ht.int32, device=cpu:0, split=None)
    >>> ht.arange(3.0)
    DNDarray([0., 1., 2.], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.arange(3, 7)
    DNDarray([3, 4, 5, 6], dtype=ht.int32, device=cpu:0, split=None)
    >>> ht.arange(3, 7, 2)
    DNDarray([3, 5], dtype=ht.int32, device=cpu:0, split=None)
    """
    num_of_param = len(args)

    # check if all positional arguments are integers
    all_ints = all([isinstance(_, int) for _ in args])

    # set start, stop, step, num according to *args
    if num_of_param == 1:
        if dtype is None:
            # use int32 as default instead of int64 used in numpy
            dtype = types.int32 if all_ints else types.float32
        start = 0
        stop = int(np.ceil(args[0]))
        step = 1
        num = stop
    elif num_of_param == 2:
        if dtype is None:
            dtype = types.int32 if all_ints else types.float32
        start = args[0]
        stop = args[1]
        step = 1
        num = int(np.ceil(stop - start))
    elif num_of_param == 3:
        if dtype is None:
            dtype = types.int32 if all_ints else types.float32
        start = args[0]
        stop = args[1]
        step = args[2]
        num = int(np.ceil((stop - start) / step))
    else:
        raise TypeError(
            f"function takes minimum one and at most 3 positional arguments ({num_of_param} given)"
        )

    # sanitize device and comm
    device = devices.sanitize_device(device)
    comm = sanitize_comm(comm)

    gshape = (num,)
    split = sanitize_axis(gshape, split)
    offset, lshape, _ = comm.chunk(gshape, split)
    balanced = True

    # compose the local tensor
    start += offset * step
    stop = start + lshape[0] * step
    data = torch.arange(start, stop, step, device=device.torch_device)

    htype = types.canonical_heat_type(dtype)
    data = data.type(htype.torch_type())

    return DNDarray(data, gshape, htype, split, device, comm, balanced)


def array(
    obj: Iterable,
    dtype: Optional[Type[datatype]] = None,
    copy: Optional[bool] = None,
    ndmin: int = 0,
    order: str = "C",
    split: Optional[int] = None,
    is_split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
) -> DNDarray:
    """
    Create a :class:`~heat.core.dndarray.DNDarray`.

    Parameters
    ----------
    obj : array_like
        A tensor or array, any object exposing the array interface, an object whose ``__array__`` method returns an
        array, or any (nested) sequence.
    dtype : datatype, optional
        The desired data-type for the array. If not given, then the type will be determined as the minimum type required
        to hold the objects in the sequence. This argument can only be used to ‘upcast’ the array. For downcasting, use
        the :func:`~heat.core.dndarray.astype` method.
    copy : bool, optional
        If ``True``, the input object is copied.
        If ``False``, input which supports the buffer protocol is never copied.
        If ``None`` (default), the function reuses the existing memory buffer if possible, and copies otherwise.
    ndmin : int, optional
        Specifies the minimum number of dimensions that the resulting array should have. Ones will, if needed, be
        attached to the shape if ``ndim > 0`` and prefaced in case of ``ndim < 0`` to meet the requirement.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).
    split : int or None, optional
        The axis along which the passed array content ``obj`` is split and distributed in memory. Mutually exclusive
        with ``is_split``.
    is_split : int or None, optional
        Specifies the axis along which the local data portions, passed in obj, are split across all machines. Useful for
        interfacing with other distributed-memory code. The shape of the global array is automatically inferred.
        Mutually exclusive with ``split``.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on (i.e. globally set default
        device).
    comm : Communication, optional
        Handle to the nodes holding distributed array chunks.

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.
    ValueError
        If ``copy`` is False but a copy is necessary to satisfy other requirements (e.g. different dtype, device, etc.).
    TypeError
        If the input object cannot be converted to a torch.Tensor, hence it cannot be converted to a :class:`~heat.core.dndarray.DNDarray`.

    Examples
    --------
    >>> ht.array([1, 2, 3])
    DNDarray([1, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    >>> ht.array([1, 2, 3.0])
    DNDarray([1., 2., 3.], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.array([[1, 2], [3, 4]])
    DNDarray([[1, 2],
              [3, 4]], dtype=ht.int64, device=cpu:0, split=None)
    >>> ht.array([1, 2, 3], ndmin=2)
    DNDarray([[1],
              [2],
              [3]], dtype=ht.int64, device=cpu:0, split=None)
    >>> ht.array([1, 2, 3], dtype=float)
    DNDarray([1., 2., 3.], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.array([1, 2, 3, 4], split=0)
    DNDarray([1, 2, 3, 4], dtype=ht.int64, device=cpu:0, split=0)
    >>> if ht.MPI_WORLD.rank == 0
    >>>     a = ht.array([1, 2], is_split=0)
    >>> else:
    >>>     a = ht.array([3, 4], is_split=0)
    >>> a
    DNDarray([1, 2, 3, 4], dtype=ht.int64, device=cpu:0, split=0)
    >>> a = np.arange(2 * 3).reshape(2, 3)
    >>> a
    array([[ 0,  1,  2],
           [ 3,  4,  5]])
    >>> a.strides
    (24, 8)
    >>> b = ht.array(a)
    >>> b
    DNDarray([[0, 1, 2],
              [3, 4, 5]], dtype=ht.int64, device=cpu:0, split=None)
    >>> b.strides
    (24, 8)
    >>> b.larray.storage()
     0
     1
     2
     3
     4
     5
    [torch.LongStorage of size 6]
    >>> c = ht.array(a, order='F')
    >>> c
    DNDarray([[0, 1, 2],
              [3, 4, 5]], dtype=ht.int64, device=cpu:0, split=None)
    >>> c.strides
    (8, 16)
    >>> c.larray.storage()
     0
     3
     1
     4
     2
     5
    [torch.LongStorage of size 6]
    >>> a = np.arange(4 * 3).reshape(4, 3)
    >>> a.strides
    (24, 8)
    >>> b = ht.array(a, order='F', split=0)
    >>> b
    DNDarray([[ 0,  1,  2],
              [ 3,  4,  5],
              [ 6,  7,  8],
              [ 9, 10, 11]], dtype=ht.int64, device=cpu:0, split=0)
    >>> b.strides
    [0/2] (8, 16)
    [1/2] (8, 16)
    >>> b.larray.storage()
    [0/2] 0
          3
          1
          4
          2
          5
         [torch.LongStorage of size 6]
    [1/2] 6
          9
          7
          10
          8
          11
         [torch.LongStorage of size 6]
    """
    # array already exists; no copy
    if isinstance(obj, DNDarray):
        if not copy:
            if (
                (dtype is None or dtype == obj.dtype)
                and (split is None or split == obj.split)
                and (is_split is None or is_split == obj.split)
                and (device is None or device == obj.device)
            ):
                return obj
        # extract the internal tensor
        obj = obj.larray

    # sanitize the data type
    if dtype is not None:
        dtype = types.canonical_heat_type(dtype)

    # sanitize device
    if device is not None:
        device = devices.sanitize_device(device)

    # initialize the array
    if bool(copy):
        if isinstance(obj, torch.Tensor):
            # TODO: watch out. At the moment clone() implies losing the underlying memory layout.
            # pytorch fix in progress
            obj = obj.clone().detach()
        else:
            try:
                obj = torch.tensor(
                    obj,
                    device=device.torch_device
                    if device is not None
                    else devices.get_device().torch_device,
                )
            except RuntimeError:
                raise TypeError("invalid data of type {}".format(type(obj)))
    else:
        if copy is False and not np.isscalar(obj) and not isinstance(obj, (Tuple, List)):
            # Python array-API compliance, cf. https://data-apis.org/array-api/2022.12/API_specification/generated/array_api.asarray.html
            if not (
                (dtype is None or dtype == types.canonical_heat_type(obj.dtype))
                and (
                    device is None
                    or device.torch_device
                    == str(getattr(obj, "device", devices.get_device().torch_device))
                )
            ):
                raise ValueError(
                    "argument `copy` is set to False, but copy of input object is necessary. \n Set copy=None to reuse the memory buffer whenever possible and allow for copies otherwise."
                )
        try:
            obj = torch.as_tensor(
                obj,
                device=device.torch_device
                if device is not None
                else devices.get_device().torch_device,
            )
        except RuntimeError:
            raise TypeError("invalid data of type {}".format(type(obj)))

    # infer dtype from obj if not explicitly given
    if dtype is None:
        dtype = types.canonical_heat_type(obj.dtype)
    else:
        torch_dtype = dtype.torch_type()
        if obj.dtype != torch_dtype:
            obj = obj.type(torch_dtype)

    # infer device from obj if not explicitly given
    if device is None:
        device = devices.sanitize_device(obj.device.type)

    if str(obj.device) != device.torch_device:
        warnings.warn(
            f"Array 'obj' is not on device '{device}'. It will be moved to it.",
            UserWarning,
        )
        obj = obj.to(device.torch_device)

    # sanitize minimum number of dimensions
    if not isinstance(ndmin, int):
        raise TypeError(f"expected ndmin to be int, but was {type(ndmin)}")

    # reshape the object to encompass additional dimensions
    ndmin_abs = abs(ndmin) - len(obj.shape)
    if ndmin_abs > 0 and ndmin > 0:
        obj = obj.reshape(obj.shape + ndmin_abs * (1,))
    if ndmin_abs > 0 > ndmin:
        obj = obj.reshape(ndmin_abs * (1,) + obj.shape)

    # sanitize the split axes, ensure mutual exclusiveness
    split = sanitize_axis(obj.shape, split)
    is_split = sanitize_axis(obj.shape, is_split)
    if split is not None and is_split is not None:
        raise ValueError("split and is_split are mutually exclusive parameters")

    # sanitize comm object
    comm = sanitize_comm(comm)

    # determine the local and the global shape. If split is None, they are identical
    gshape = list(obj.shape)
    lshape = gshape.copy()
    balanced = True

    # content shall be split, chunk the passed data object up
    if comm.size == 1 or split is None and is_split is None:
        obj = sanitize_memory_layout(obj, order=order)
        split = is_split if is_split is not None else split

    elif split is not None:
        # only keep local slice
        _, _, slices = comm.chunk(gshape, split)
        _ = obj[slices].clone()
        del obj

        obj = sanitize_memory_layout(_, order=order)

    # check with the neighboring rank whether the local shape would fit into a global shape
    elif is_split is not None:
        obj = sanitize_memory_layout(obj, order=order)

        # Check whether the shape of distributed data
        # matches in all dimensions except the split axis
        neighbour_shape = np.array(gshape)
        lshape = np.array(lshape)

        if comm.rank < comm.size - 1:
            comm.Isend(lshape, dest=comm.rank + 1)
        if comm.rank != 0:
            # look into the message of the neighbor to see whether the shape length fits
            status = MPI.Status()
            comm.Probe(source=comm.rank - 1, status=status)
            length = status.Get_count() // lshape.dtype.itemsize
            del status
            # the number of shape elements does not match with the 'left' rank
            if length != len(lshape):
                discard_buffer = np.empty(length)
                comm.Recv(discard_buffer, source=comm.rank - 1)
                lshape[is_split] = np.iinfo(lshape.dtype).min
            else:
                # check whether the individual shape elements match
                comm.Recv(neighbour_shape, source=comm.rank - 1)
                for i in range(length):
                    if i != is_split:
                        if lshape[i] != neighbour_shape[i]:
                            lshape[is_split] = np.iinfo(lshape.dtype).min
            del neighbour_shape

        # sum up the elements along the split dimension
        reduction_buffer = np.array(lshape[is_split])
        comm.Allreduce(MPI.IN_PLACE, reduction_buffer, MPI.MIN)
        if reduction_buffer < 0:
            raise ValueError(
                "Unable to construct DNDarray. Local data slices have inconsistent shapes or dimensions."
            )

        total_split_shape = np.array(lshape[is_split])
        comm.Allreduce(MPI.IN_PLACE, total_split_shape, MPI.SUM)
        gshape[is_split] = total_split_shape.item()
        split = is_split
        # compare to calculated balanced lshape (cf. dndarray.is_balanced())
        _, test_lshape, _ = comm.chunk(gshape, split)
        match = (test_lshape == lshape).all().astype(int)
        gmatch = comm.allreduce(match, MPI.SUM)
        if gmatch != comm.size:
            balanced = False

    return DNDarray(obj, tuple(gshape), dtype, split, device, comm, balanced)


def asarray(
    obj: Iterable,
    dtype: Optional[Type[datatype]] = None,
    copy: Optional[bool] = None,
    order: str = "C",
    is_split: Optional[bool] = None,
    device: Optional[Union[str, Device]] = None,
) -> DNDarray:
    """
    Convert ``obj`` to a DNDarray. If ``obj`` is a `DNDarray` or `Tensor` with the same `dtype` and `device` or if the
    data is an `ndarray` of the corresponding ``dtype`` and the ``device`` is the CPU, no copy will be performed.

    Parameters
    ----------
    obj : iterable
        Input data, in any form that can be converted to an array. This includes e.g. lists, lists of tuples, tuples,
        tuples of tuples, tuples of lists and ndarrays.
    dtype : dtype, optional
        By default, the data-type is inferred from the input data.
    copy : bool, optional
        If ``True``, then the object is copied.  If ``False``, the object is not copied and a ``ValueError`` is
        raised in the case a copy would be necessary. If ``None``, a copy will only be made if `obj` is a nested
        sequence or if a copy is needed to satisfy any of the other requirements, e.g. ``dtype``.
    order: str, optional
        Whether to use row-major (C-style) or column-major (Fortran-style) memory representation. Defaults to ‘C’.
    is_split : None or int, optional
        Specifies the axis along which the local data portions, passed in obj, are split across all MPI processes. Useful for
        interfacing with other HPC code. The shape of the global tensor is automatically inferred.
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on. By default, it is inferred from the input data.

    Examples
    --------
    >>> a = [1,2]
    >>> ht.asarray(a)
    DNDarray([1, 2], dtype=ht.int64, device=cpu:0, split=None)
    >>> a = np.array([1,2,3])
    >>> n = ht.asarray(a)
    >>> n
    DNDarray([1, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    >>> n[0] = 0
    >>> a
    DNDarray([0, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    >>> a = torch.tensor([1,2,3])
    >>> t = ht.asarray(a)
    >>> t
    DNDarray([1, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    >>> t[0] = 0
    >>> a
    DNDarray([0, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    >>> a = ht.array([1,2,3,4], dtype=ht.float32)
    >>> ht.asarray(a, dtype=ht.float32) is a
    True
    >>> ht.asarray(a, dtype=ht.float64) is a
    False
    """
    return array(obj, dtype=dtype, copy=copy, order=order, is_split=is_split, device=device)


def empty(
    shape: Union[int, Sequence[int]],
    dtype: Type[datatype] = types.float32,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Returns a new uninitialized :class:`~heat.core.dndarray.DNDarray` of given shape and data type. May be allocated
    split up across multiple nodes along the specified axis.

    Parameters
    ----------
    shape : int or Sequence[int,...]
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : datatype
        The desired HeAT data type for the array.
    split: int, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device`. the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> ht.empty(3)
    DNDarray([0., 0., 0.], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.empty(3, dtype=ht.int)
    DNDarray([59140784,        0, 59136816], dtype=ht.int32, device=cpu:0, split=None)
    >>> ht.empty((2, 3,))
    DNDarray([[-1.7206e-10,  4.5905e-41, -1.7206e-10],
              [ 4.5905e-41,  4.4842e-44,  0.0000e+00]], dtype=ht.float32, device=cpu:0, split=None)
    """
    # TODO: implement 'K' option when torch.clone() fix to preserve memory layout is released.
    return __factory(shape, dtype, split, torch.empty, device, comm, order)


def empty_like(
    a: DNDarray,
    dtype: Optional[Type[datatype]] = None,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Returns a new uninitialized :class:`~heat.core.dndarray.DNDarray` with the same type, shape and data distribution
    of given object. Data type and data distribution strategy can be explicitly overriden.

    Parameters
    ----------
    a : DNDarray
        The shape and data-type of ``a`` define these same attributes of the returned array. Uninitialized array with
        the same shape, type and split axis as ``a`` unless overriden.
    dtype : datatype, optional
        Overrides the data type of the result.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> x = ht.ones((2, 3,))
    >>> x
    DNDarray([[1., 1., 1.],
              [1., 1., 1.]], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.empty_like(x)
    DNDarray([[-1.7205e-10,  4.5905e-41,  7.9442e-37],
              [ 0.0000e+00,  4.4842e-44,  0.0000e+00]], dtype=ht.float32, device=cpu:0, split=None)
    """
    return __factory_like(a, dtype, split, empty, device, comm, order=order)


def eye(
    shape: Union[int, Sequence[int]],
    dtype: Type[datatype] = types.float32,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Returns a new 2-D :class:`~heat.core.dndarray.DNDarray` with ones on the diagonal and zeroes elsewhere, i.e. an
    identity matrix.

    Parameters
    ----------
    shape : int or Sequence[int,...]
        The shape of the data-type. If only one number is provided, returning array will be square with that size. In
        other cases, the first value represents the number rows, the second the number of columns.
    dtype : datatype, optional
        Overrides the data type of the result.
    split : int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm : Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> ht.eye(2)
    DNDarray([[1., 0.],
              [0., 1.]], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.eye((2, 3), dtype=ht.int32)
    DNDarray([[1, 0, 0],
              [0, 1, 0]], dtype=ht.int32, device=cpu:0, split=None)
    """
    # TODO: implement 'K' option when torch.clone() fix to preserve memory layout is released.
    # Determine the actual size of the resulting data
    gshape = shape
    if isinstance(gshape, int):
        gshape = (gshape, gshape)
    if len(gshape) == 1:
        gshape = gshape * 2

    split = sanitize_axis(gshape, split)
    device = devices.sanitize_device(device)
    comm = sanitize_comm(comm)
    offset, lshape, _ = comm.chunk(gshape, split)
    balanced = True

    # start by creating tensor filled with zeroes
    data = torch.zeros(
        lshape, dtype=types.canonical_heat_type(dtype).torch_type(), device=device.torch_device
    )

    # insert ones at the correct positions
    for i in range(min(lshape)):
        pos_x = i if split == 0 else i + offset
        pos_y = i if split == 1 else i + offset
        if pos_x >= lshape[0] or pos_y >= lshape[1]:
            break
        data[pos_x][pos_y] = 1

    data = sanitize_memory_layout(data, order=order)

    return DNDarray(
        data, gshape, types.canonical_heat_type(data.dtype), split, device, comm, balanced
    )


def __factory(
    shape: Union[int, Sequence[int]],
    dtype: Type[datatype],
    split: Optional[int],
    local_factory: Callable,
    device: Device,
    comm: Communication,
    order: str,
) -> DNDarray:
    """
    Abstracted factory function for HeAT :class:`~heat.core.dndarray.DNDarray` initialization.

    Parameters
    ----------
    shape : int or Sequence[ints,...]
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : datatype
        The desired HeAT data type for the array, defaults to ht.float32.
    split : int or None
        The axis along which the array is split and distributed.
    local_factory : callable
        Function that creates the local PyTorch tensor for the DNDarray.
    device : Device
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm : Communication
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.
    """
    # clean the user input
    shape = sanitize_shape(shape)
    dtype = types.canonical_heat_type(dtype)
    split = sanitize_axis(shape, split)
    device = devices.sanitize_device(device)
    comm = sanitize_comm(comm)

    # chunk the shape if necessary
    _, local_shape, _ = comm.chunk(shape, split)

    # create the torch data using the factory function
    data = local_factory(local_shape, dtype=dtype.torch_type(), device=device.torch_device)
    data = sanitize_memory_layout(data, order=order)

    return DNDarray(data, shape, dtype, split, device, comm, balanced=True)


def __factory_like(
    a: DNDarray,
    dtype: Type[datatype],
    split: Optional[int],
    factory: Callable,
    device: Device,
    comm: Communication,
    order: str = "C",
    **kwargs,
) -> DNDarray:
    """
    Abstracted '...-like' factory function for HeAT :class:`~heat.core.dndarray.DNDarray` initialization

    Parameters
    ----------
    a : DNDarray
        The shape and data-type of ``a`` define these same attributes of the returned array.
    dtype : datatype
        The desired HeAT data type for the array.
    split: int or None, optional
        The axis along which the array is split and distributed, defaults to no distribution).
    factory : function
        Function that creates a DNDarray.
    device : str
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.
    """
    # TODO: implement 'K' option when torch.clone() fix to preserve memory layout is released.
    # determine the global shape of the object to create
    # attempt in this order: shape property, length of object or default shape (1,)
    try:
        shape = a.shape
    except AttributeError:
        try:
            shape = (len(a),)
        except TypeError:
            shape = (1,)

    # infer the data type, otherwise default to float32
    if dtype is None:
        try:
            dtype = types.heat_type_of(a)
        except TypeError:
            dtype = types.float32

    # infer split axis
    if split is None:
        try:
            split = None if isinstance(a, str) else a.split
        except AttributeError:
            # do not split at all
            pass

    # use the default communicator, if not set
    comm = sanitize_comm(comm)

    return factory(shape, dtype=dtype, split=split, device=device, comm=comm, order=order, **kwargs)


def from_partitioned(x, comm: Optional[Communication] = None) -> DNDarray:
    """
    Return a newly created DNDarray constructed from the '__partitioned__' attributed of the input object.
    Memory of local partitions will be shared (zero-copy) as long as supported by data objects.
    Currently supports numpy ndarrays and torch tensors as data objects.
    Current limitations:
      * Partitions must be ordered in the partition-grid by rank
      * Only one split-axis
      * Only one partition per rank
      * Only SPMD-style __partitioned__

    Parameters
    ----------
    x : object
        Requires x.__partitioned__
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.

    See also
    --------
    :func:`ht.core.DNDarray.create_partition_interface <ht.core.DNDarray.create_partition_interface>`.

    Raises
    ------
    AttributeError
        If not hasattr(x, "__partitioned__") or if underlying data has no dtype.
    TypeError
        If it finds an unsupported array types
    RuntimeError
        If other unsupported content is found.

    Examples
    --------
    >>> import heat as ht
    >>> a = ht.ones((44,55), split=0)
    >>> b = ht.from_partitioned(a)
    >>> assert (a==b).all()
    >>> a[40] = 4711
    >>> assert (a==b).all()
    """
    comm = sanitize_comm(comm)
    parted = x.__partitioned__
    return __from_partition_dict_helper(parted, comm)


def from_partition_dict(parted: dict, comm: Optional[Communication] = None) -> DNDarray:
    """
    Return a newly created DNDarray constructed from the '__partitioned__' attributed of the input object.
    Memory of local partitions will be shared (zero-copy) as long as supported by data objects.
    Currently supports numpy ndarrays and torch tensors as data objects.
    Current limitations:
      * Partitions must be ordered in the partition-grid by rank
      * Only one split-axis
      * Only one partition per rank
      * Only SPMD-style __partitioned__

    Parameters
    ----------
    parted : dict
        A partition dictionary used to create the new DNDarray
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.

    See also
    --------
    :func:`ht.core.DNDarray.create_partition_interface <ht.core.DNDarray.create_partition_interface>`.

    Raises
    ------
    AttributeError
        If not hasattr(x, "__partitioned__") or if underlying data has no dtype.
    TypeError
        If it finds an unsupported array types
    RuntimeError
        If other unsupported content is found.

    Examples
    --------
    >>> import heat as ht
    >>> a = ht.ones((44,55), split=0)
    >>> b = ht.from_partition_dict(a.__partitioned__)
    >>> assert (a==b).all()
    >>> a[40] = 4711
    >>> assert (a==b).all()
    """
    comm = sanitize_comm(comm)
    return __from_partition_dict_helper(parted, comm)


def __from_partition_dict_helper(parted: dict, comm: Communication):
    # helper to create a DNDarray from a partition table (dictionary)
    # the dictionary must be in the same form as the DNDarray.__partitioned__ property creates
    if "locals" not in parted:
        raise RuntimeError("Non-SPMD __partitioned__ not supported")
    try:
        gshape = parted["shape"]
    except KeyError:
        raise RuntimeError(
            "partition dictionary must have a 'shape' entry, see DNDarray.create_partition_interface for more details"
        )
    try:
        lparts = parted["locals"]
    except KeyError:
        raise RuntimeError(
            "partition dictionary must have a 'local' entry, see DNDarray.create_partition_interface for more details"
        )
    if len(lparts) != 1:
        raise RuntimeError("Only exactly one partition per rank supported (yet)")
    parts = parted["partitions"]
    lpart = parted["get"](parts[lparts[0]]["data"])
    if isinstance(lpart, np.ndarray):
        data = torch.from_numpy(lpart)
    elif isinstance(lpart, torch.Tensor):
        data = lpart
    else:
        raise TypeError(f"Only numpy arrays and torch tensors supported (not {type(lpart)}")
    htype = types.canonical_heat_type(data.dtype)

    # get split axis
    gshape_list = list(gshape)
    lshape_list = list(data.shape)
    shape_diff = torch.tensor(
        [g - l for g, l in zip(gshape_list, lshape_list)]
    )  # dont care about device
    nz = torch.nonzero(shape_diff)

    if nz.numel() > 1:
        raise RuntimeError("only one split axis allowed, check the ")
    elif nz.numel() == 1:
        split = nz[0].item()
    else:
        split = None

    expected = {
        int(x["location"][0]): (
            comm.chunk(gshape, split, x["location"][0])[1:],
            (x["shape"], x["start"]),
        )
        for x in parts.values()
    }
    balanced = all(x[0][0] == x[1][0] for x in expected.values())

    ret = DNDarray(
        data, gshape, htype, split, devices.sanitize_device(None), sanitize_comm(comm), balanced
    )
    ret.__partitions_dict__ = parted

    return ret


def full(
    shape: Union[int, Sequence[int]],
    fill_value: Union[int, float],
    dtype: Type[datatype] = types.float32,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Return a new :class:`~heat.core.dndarray.DNDarray` of given shape and type, filled with ``fill_value``.

    Parameters
    ----------
    shape : int or Sequence[int,...]
        Shape of the new array, e.g., (2, 3) or 2.
    fill_value : scalar
        Fill value.
    dtype : datatype, optional
        The desired data-type for the array
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> ht.full((2, 2), ht.inf)
    DNDarray([[inf, inf],
              [inf, inf]], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.full((2, 2), 10)
    DNDarray([[10., 10.],
              [10., 10.]], dtype=ht.float32, device=cpu:0, split=None)
    """

    def local_factory(*args, **kwargs):
        return torch.full(*args, fill_value=fill_value, **kwargs)

    # Will be redundant with PyTorch 1.7
    if isinstance(fill_value, complex):
        dtype = types.complex64

    return __factory(shape, dtype, split, local_factory, device, comm, order=order)


def full_like(
    a: DNDarray,
    fill_value: Union[int, float],
    dtype: Type[datatype] = types.float32,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Return a full :class:`~heat.core.dndarray.DNDarray` with the same shape and type as a given array.

    Parameters
    ----------
    a : DNDarray
        The shape and data-type of ``a`` define these same attributes of the returned array.
    fill_value : scalar
        Fill value.
    dtype : datatype, optional
        Overrides the data type of the result.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> x = ht.zeros((2, 3,))
    >>> x
    DNDarray([[0., 0., 0.],
              [0., 0., 0.]], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.full_like(x, 1.0)
    DNDarray([[1., 1., 1.],
              [1., 1., 1.]], dtype=ht.float32, device=cpu:0, split=None)
    """
    return __factory_like(a, dtype, split, full, device, comm, fill_value=fill_value, order=order)


def linspace(
    start: Union[int, float],
    stop: Union[int, float],
    num: int = 50,
    endpoint: bool = True,
    retstep: bool = False,
    dtype: Optional[Type[datatype]] = None,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
) -> Tuple[DNDarray, float]:
    """
    Returns num evenly spaced samples, calculated over the interval ``[start, stop]``. The endpoint of the interval can
    optionally be excluded. There are num equally spaced samples in the closed interval ``[start, stop]`` or the
    half-open interval ``[start, stop)`` (depending on whether endpoint is ``True`` or ``False``).

    Parameters
    ----------
    start: scalar or scalar-convertible
        The starting value of the sample interval, maybe a sequence if convertible to scalar
    stop: scalar or scalar-convertible
        The end value of the sample interval, unless is set to False. In that case, the sequence consists of all but the
        last of ``num+1`` evenly spaced samples, so that stop is excluded. Note that the step size changes when endpoint
        is ``False``.
    num: int, optional
        Number of samples to generate, defaults to 50. Must be non-negative.
    endpoint: bool, optional
        If ``True``, stop is the last sample, otherwise, it is not included.
    retstep: bool, optional
        If ``True``, return (samples, step), where step is the spacing between samples.
    dtype: dtype, optional
        The type of the output array.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm : Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.

    Examples
    --------
    >>> ht.linspace(2.0, 3.0, num=5)
    DNDarray([2.0000, 2.2500, 2.5000, 2.7500, 3.0000], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.linspace(2.0, 3.0, num=5, endpoint=False)
    DNDarray([2.0000, 2.2000, 2.4000, 2.6000, 2.8000], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.linspace(2.0, 3.0, num=5, retstep=True)
    (DNDarray([2.0000, 2.2500, 2.5000, 2.7500, 3.0000], dtype=ht.float32, device=cpu:0, split=None), 0.25)
    """
    # sanitize input parameters
    start = float(start)
    stop = float(stop)
    num = int(num)
    if num < 0:
        raise ValueError(f"number of samples 'num' must be non-negative integer, but was {num}")
    step = (stop - start) / max(1, num - 1 if endpoint else num)

    # sanitize device and comm
    device = devices.sanitize_device(device)
    comm = sanitize_comm(comm)

    # infer local and global shapes
    gshape = (num,)
    split = sanitize_axis(gshape, split)
    offset, lshape, _ = comm.chunk(gshape, split)
    balanced = True

    # compose the local tensor
    start += offset * step
    stop = start + lshape[0] * step - step
    data = torch.linspace(start, stop, lshape[0], device=device.torch_device)
    if dtype is not None:
        data = data.type(types.canonical_heat_type(dtype).torch_type())

    # construct the resulting global tensor
    ht_tensor = DNDarray(
        data, gshape, types.canonical_heat_type(data.dtype), split, device, comm, balanced
    )

    if retstep:
        return ht_tensor, step
    return ht_tensor


def logspace(
    start: Union[int, float],
    stop: Union[int, float],
    num: int = 50,
    endpoint: bool = True,
    base: float = 10.0,
    dtype: Optional[Type[datatype]] = None,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
) -> DNDarray:
    """
    Return numbers spaced evenly on a log scale. In linear space, the sequence starts at ``base**start`` (``base`` to
    the power of ``start``) and ends with ``base**stop`` (see ``endpoint`` below).

    Parameters
    ----------
    start : scalar or scalar-convertible
        ``base**start`` is the starting value of the sequence.
    stop : scalar or scalar-convertible
        ``base**stop`` is the final value of the sequence, unless `endpoint` is ``False``.  In that case, ``num+1``
        values are spaced over the interval in log-space, of which all but the last (a sequence of length ``num``) are
        returned.
    num : int, optional
        Number of samples to generate.
    endpoint : bool, optional
        If ``True``, `stop` is the last sample. Otherwise, it is not included.
    base : float, optional
        The base of the log space. The step size between the elements in :math:`ln(samples) / ln(base)` (or
        :math:`base(samples)`) is uniform.
    dtype : datatype, optional
        The type of the output array.  If ``dtype`` is not given, infer the data type from the other input arguments.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.

    See Also
    --------
    :func:`arange` : Similar to :func:`linspace`, with the step size specified instead of the
        number of samples. Note that, when used with a float endpoint, the endpoint may or may not be included.

    :func:`linspace` : Similar to ``logspace``, but with the samples uniformly distributed in linear space, instead of
        log space.

    Examples
    --------
    >>> ht.logspace(2.0, 3.0, num=4)
    DNDarray([ 100.0000,  215.4434,  464.1590, 1000.0000], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.logspace(2.0, 3.0, num=4, endpoint=False)
    DNDarray([100.0000, 177.8279, 316.2278, 562.3413], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.logspace(2.0, 3.0, num=4, base=2.0)
    DNDarray([4.0000, 5.0397, 6.3496, 8.0000], dtype=ht.float32, device=cpu:0, split=None)
    """
    y = linspace(start, stop, num=num, endpoint=endpoint, split=split, device=device, comm=comm)
    if dtype is None:
        return pow(base, y)
    return pow(base, y).astype(dtype, copy=False)


def meshgrid(*arrays: Sequence[DNDarray], indexing: str = "xy") -> List[DNDarray]:
    """
    Returns coordinate matrices from coordinate vectors.

    Parameters
    ----------
    arrays : Sequence[ DNDarray ]
        one-dimensional arrays representing grid coordinates. If exactly one vector is distributed, the returned matrices will
        be distributed along the axis equal to the index of this vector in the input list.
    indexing : str, optional
        Cartesian ‘xy’ or matrix ‘ij’ indexing of output. It is ignored if zero or one one-dimensional arrays are provided. Default: 'xy' .

    Raises
    ------
    ValueError
        If `indexing` is not 'xy' or 'ij'.
    ValueError
        If more than one input vector is distributed.

    Examples
    --------
    >>> x = ht.arange(4)
    >>> y = ht.arange(3)
    >>> xx, yy = ht.meshgrid(x,y)
    >>> xx
    DNDarray([[0, 1, 2, 3],
          [0, 1, 2, 3],
          [0, 1, 2, 3]], dtype=ht.int32, device=cpu:0, split=None)
    >>> yy
    DNDarray([[0, 0, 0, 0],
          [1, 1, 1, 1],
          [2, 2, 2, 2]], dtype=ht.int32, device=cpu:0, split=None)
    """
    splitted = None

    if indexing not in ["xy", "ij"]:
        raise ValueError("Valid values for `indexing` are 'xy' and 'ij'.")

    if len(arrays) == 0:
        return []

    arrays = sanitize_sequence(arrays)

    for idx, array in enumerate(arrays):
        sanitize_in(array)
        if array.split is not None:
            if splitted is not None:
                raise ValueError("split != None are not supported.")
            splitted = idx

    # pytorch does not support the indexing keyword: switch vectors
    if indexing == "xy" and len(arrays) > 1:
        arrays[0], arrays[1] = arrays[1], arrays[0]
        if splitted == 0:
            arrays[0] = arrays[0].resplit(0)
            arrays[1] = arrays[1].resplit(None)
        elif splitted == 1:
            arrays[0] = arrays[0].resplit(None)
            arrays[1] = arrays[1].resplit(0)

    grids = torch.meshgrid(*(array.larray for array in arrays))

    # pytorch does not support indexing keyword: switch back
    if indexing == "xy" and len(arrays) > 1:
        grids = list(grids)
        grids[0], grids[1] = grids[1], grids[0]

    shape = tuple(array.size for array in arrays)

    return [
        DNDarray(
            array=grid,
            gshape=shape,
            dtype=types.heat_type_of(grid),
            split=splitted,
            device=devices.sanitize_device(grid.device.type),
            comm=sanitize_comm(None),
            balanced=True,
        )
        for grid in grids
    ]


def ones(
    shape: Union[int, Sequence[int]],
    dtype: Type[datatype] = types.float32,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Returns a new :class:`~heat.core.dndarray.DNDarray` of given shape and data type filled with one. May be allocated
    split up across multiple nodes along the specified axis.

    Parameters
    ----------
    shape : int or Sequence[int,...]
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : datatype, optional
        The desired HeAT data type for the array.
    split : int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm : Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> ht.ones(3)
    DNDarray([1., 1., 1.], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.ones(3, dtype=ht.int)
    DNDarray([1, 1, 1], dtype=ht.int32, device=cpu:0, split=None)
    >>> ht.ones((2, 3,))
    DNDarray([[1., 1., 1.],
          [1., 1., 1.]], dtype=ht.float32, device=cpu:0, split=None)
    """
    # TODO: implement 'K' option when torch.clone() fix to preserve memory layout is released.
    return __factory(shape, dtype, split, torch.ones, device, comm, order)


def ones_like(
    a: DNDarray,
    dtype: Optional[Type[datatype]] = None,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Returns a new :class:`~heat.core.dndarray.DNDarray` filled with ones with the same type,
    shape and data distribution of given object. Data type and data distribution strategy can be explicitly overriden.

    Parameters
    ----------
    a : DNDarray
        The shape and data-type of ``a`` define these same attributes of the returned array.
    dtype : datatype, optional
        Overrides the data type of the result.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> x = ht.zeros((2, 3,))
    >>> x
    DNDarray([[0., 0., 0.],
              [0., 0., 0.]], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.ones_like(x)
    DNDarray([[1., 1., 1.],
              [1., 1., 1.]], dtype=ht.float32, device=cpu:0, split=None)
    """
    return __factory_like(a, dtype, split, ones, device, comm, order=order)


def zeros(
    shape: Union[int, Sequence[int]],
    dtype: Type[datatype] = types.float32,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Returns a new :class:`~heat.core.dndarray.DNDarray` of given shape and data type filled with zero values.
    May be allocated split up across multiple nodes along the specified axis.

    Parameters
    ----------
    shape : int or Sequence[int,...]
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : datatype
        The desired HeAT data type for the array.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> ht.zeros(3)
    DNDarray([0., 0., 0.], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.zeros(3, dtype=ht.int)
    DNDarray([0, 0, 0], dtype=ht.int32, device=cpu:0, split=None)
    >>> ht.zeros((2, 3,))
    DNDarray([[0., 0., 0.],
              [0., 0., 0.]], dtype=ht.float32, device=cpu:0, split=None)
    """
    # TODO: implement 'K' option when torch.clone() fix to preserve memory layout is released.
    return __factory(shape, dtype, split, torch.zeros, device, comm, order=order)


def zeros_like(
    a: DNDarray,
    dtype: Optional[Type[datatype]] = None,
    split: Optional[int] = None,
    device: Optional[Device] = None,
    comm: Optional[Communication] = None,
    order: str = "C",
) -> DNDarray:
    """
    Returns a new :class:`~heat.core.dndarray.DNDarray` filled with zeros with the same type, shape and data
    distribution of given object. Data type and data distribution strategy can be explicitly overriden.

    Parameters
    ----------
    a : DNDarray
        The shape and data-type of ``a`` define these same attributes of the returned array.
    dtype : datatype, optional
        Overrides the data type of the result.
    split: int or None, optional
        The axis along which the array is split and distributed; ``None`` means no distribution.
    device : str or Device, optional
        Specifies the :class:`~heat.core.devices.Device` the array shall be allocated on, defaults to globally set
        default device.
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this array.
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created array. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).

    Raises
    ------
    NotImplementedError
        If order is one of the NumPy options ``'K'`` or ``'A'``.

    Examples
    --------
    >>> x = ht.ones((2, 3,))
    >>> x
    DNDarray([[1., 1., 1.],
              [1., 1., 1.]], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.zeros_like(x)
    DNDarray([[0., 0., 0.],
              [0., 0., 0.]], dtype=ht.float32, device=cpu:0, split=None)
    """
    # TODO: implement 'K' option when torch.clone() fix to preserve memory layout is released.
    return __factory_like(a, dtype, split, zeros, device, comm, order=order)



"""
implementations of the different dtypes supported in heat and the
"""

from __future__ import annotations

import builtins
import collections
import numpy as np
import torch

from . import communication
from . import devices
from . import factories
from . import _operations
from . import sanitation

from typing import Type, Union, Tuple, Any, Iterable, Optional

__all__ = [
    "datatype",
    "number",
    "integer",
    "signedinteger",
    "unsignedinteger",
    "bool",
    "bool_",
    "floating",
    "int8",
    "byte",
    "int16",
    "short",
    "int32",
    "int",
    "int64",
    "long",
    "uint8",
    "ubyte",
    "float32",
    "float",
    "float_",
    "float64",
    "double",
    "flexible",
    "can_cast",
    "canonical_heat_type",
    "heat_type_is_exact",
    "heat_type_is_inexact",
    "iscomplex",
    "isreal",
    "issubdtype",
    "heat_type_of",
    "promote_types",
    "result_type",
    "complex64",
    "cfloat",
    "csingle",
    "complex128",
    "cdouble",
]
__api__ = []


class datatype:
    """
    Defines the basic heat data types in the hierarchy as shown below. Design inspired by the Python package numpy.
    As part of the type-hierarchy: xx -- is bit-width \n
        - generic \n
            - bool, bool_ (kind=?) \n
            - number \n
                - integer \n
                    - signedinteger (intxx)(kind=b, i) \n
                        - int8, byte \n
                        - int16, short \n
                        - int32, int \n
                        - int64, long
                    - unsignedinteger (uintxx)(kind=B, u) \n
                        - uint8, ubyte
                - floating (floatxx) (kind=f) \n
                    - float32, float, float_ \n
                    - float64, double (double)
            - flexible (currently unused, placeholder for characters) \n
    """

    def __new__(
        cls,
        *value,
        device: Optional[Union[str, devices.Device]] = None,
        comm: Optional[communication.Communication] = None,
    ) -> dndarray.DNDarray:
        """
        Create a new DNDarray. See :func:`ht.array <heat.core.factories.array>` for more info on general
        DNDarray creation.

        Parameters
        ----------
        value: array_like
            The values for the DNDarray which will be created
        device: devices.Device
            The device on which to place the created DNDarray
        comm: communication.Communication
            The MPI communication object to use in distribution and further operations
        """
        torch_type = cls.torch_type()
        if torch_type is NotImplemented:
            raise TypeError(f"cannot create '{cls}' instances")

        value_count = len(value)

        # sanitize the distributed processing flags
        comm = communication.sanitize_comm(comm)
        device = devices.sanitize_device(device)

        # check whether there are too many arguments
        if value_count >= 2:
            raise TypeError(f"function takes at most 1 argument ({value_count} given)")
        elif value_count == 0:
            value = ((0,),)

        # otherwise, attempt to create a torch tensor of given type
        try:
            array = value[0]._DNDarray__array.type(torch_type)
            return dndarray.DNDarray(
                array,
                gshape=value[0].shape,
                dtype=cls,
                split=value[0].split,
                comm=comm,
                device=device,
                balanced=value[0].balanced,
            )
        except AttributeError:
            # this is the case of that the first/only element of value is not a DNDarray
            array = torch.tensor(*value, dtype=torch_type, device=device.torch_device)
        except TypeError as exception:
            # re-raise the exception to be consistent with numpy's exception interface
            raise ValueError(str(exception))

        return dndarray.DNDarray(
            array, tuple(array.shape), cls, split=None, device=device, comm=comm, balanced=True
        )

    @classmethod
    def torch_type(cls) -> NotImplemented:
        """
        Torch Datatype
        """
        return NotImplemented

    @classmethod
    def char(cls) -> NotImplemented:
        """
        Datatype short-hand name
        """
        return NotImplemented


class bool(datatype):
    """
    The boolean datatype in Heat
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.bool

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "u1"


class number(datatype):
    """
    The general number datatype. Integer and Float classes will inherit from this.
    """

    pass


class integer(number):
    """
    The general integer datatype. Specific integer classes inherit from this.
    """

    pass


class signedinteger(integer):
    """
    The general signed integer datatype.
    """

    pass


class int8(signedinteger):
    """
    8 bit signed integer datatype
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int8

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i1"


class int16(signedinteger):
    """
    16 bit signed integer datatype
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int16

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i2"


class int32(signedinteger):
    """
    32 bit signed integer datatype
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int32

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i4"


class int64(signedinteger):
    """
    64 bit signed integer datatype
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int64

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i8"


class unsignedinteger(integer):
    """
    The general unsigned integer datatype
    """

    pass


class uint8(unsignedinteger):
    """
    8 bit unsigned integer datatype
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.uint8

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "u1"


class floating(number):
    """
    The general floating point datatype class.
    """

    pass


class float32(floating):
    """
    The 32 bit floating point datatype
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.float32

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "f4"


class float64(floating):
    """
    The 64 bit floating point datatype
    """

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatye
        """
        return torch.float64

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "f8"


class flexible(datatype):
    """
    The general flexible datatype. Currently unused, placeholder for characters
    """

    pass


class complex(number):
    """
    The general complex datatype class.
    """

    pass


class complex64(complex):
    """
    The complex 64 bit datatype. Both real and imaginary are 32 bit floating point
    """

    @classmethod
    def torch_type(cls):
        """
        Torch Datatype
        """
        return torch.complex64

    @classmethod
    def char(cls):
        """
        Datatype short-hand name
        """
        return "c8"


class complex128(complex):
    """
    The complex 128 bit datatype. Both real and imaginary are 64 bit floating point
    """

    @classmethod
    def torch_type(cls):
        """
        Torch Datatype
        """
        return torch.complex128

    @classmethod
    def char(cls):
        """
        Datatype short-hand name
        """
        return "c16"


# definition of aliases
bool_ = bool
ubyte = uint8
byte = int8
short = int16
int = int32
int_ = int32
long = int64
float = float32
float_ = float32
double = float64
cfloat = complex64
csingle = complex64
cdouble = complex128

_complexfloating = (complex64, complex128)

_inexact = (
    # float16,
    float32,
    float64,
    *_complexfloating,
)

_exact = (uint8, int8, int16, int32, int64)

# type mappings for type strings and builtins types
__type_mappings = {
    # type strings
    "?": bool,
    "B": uint8,
    "b": int8,
    "h": int16,
    "i": int32,
    "l": int64,
    "f": float32,
    "d": float64,
    "F": complex64,
    "D": complex128,
    "b1": bool,
    "u": uint8,
    "u1": uint8,
    "i1": int8,
    "i2": int16,
    "i4": int32,
    "i8": int64,
    "f4": float32,
    "f8": float64,
    "c8": complex64,
    "c16": complex128,
    # numpy types
    np.bool_: bool,
    np.uint8: uint8,
    np.int8: int8,
    np.int16: int16,
    np.int32: int32,
    np.int64: int64,
    np.float32: float32,
    np.float64: float64,
    np.complex64: complex64,
    np.complex128: complex128,
    # torch types
    torch.bool: bool,
    torch.uint8: uint8,
    torch.int8: int8,
    torch.int16: int16,
    torch.int32: int32,
    torch.int64: int64,
    torch.float32: float32,
    torch.float64: float64,
    torch.complex64: complex64,
    torch.complex128: complex128,
    # builtins
    builtins.bool: bool,
    builtins.int: int32,
    builtins.float: float32,
    builtins.complex: complex64,
}


def canonical_heat_type(a_type: Union[str, Type[datatype], Any]) -> Type[datatype]:
    """
    Canonicalize the builtin Python type, type string or HeAT type into a canonical HeAT type.

    Parameters
    ----------
    a_type : type, str, datatype
        A description for the type. It may be a a Python builtin type, string or an HeAT type already.
        In the three former cases the according mapped type is looked up, in the latter the type is simply returned.

    Raises
    -------
    TypeError
        If the type cannot be converted.
    """
    # already a heat type
    try:
        if issubclass(a_type, datatype):
            return a_type
    except TypeError:
        pass

    # extract type of numpy.dtype
    a_type = getattr(a_type, "type", a_type)

    # try to look the corresponding type up
    try:
        return __type_mappings[a_type]
    except KeyError:
        raise TypeError(f"data type {a_type} is not understood")


def heat_type_is_exact(ht_dtype: Type[datatype]) -> bool:
    """
    Check if HeAT type is an exact type, i.e an integer type. True if ht_dtype is an integer, False otherwise

    Parameters
    ----------
    ht_dtype: Type[datatype]
        HeAT type to check
    """
    return ht_dtype in _exact


def heat_type_is_inexact(ht_dtype: Type[datatype]) -> bool:
    """
    Check if HeAT type is an inexact type, i.e floating point type. True if ht_dtype is a float, False otherwise

    Parameters
    ----------
    ht_dtype: Type[datatype]
        HeAT type to check
    """
    return ht_dtype in _inexact


def heat_type_is_complexfloating(ht_dtype: Type[datatype]) -> bool:
    """
    Check if HeAT type is a complex floading point number, i.e complex64

    Parameters
    ----------
    ht_dtype: ht.dtype
        HeAT type to check

    Returns
    -------
    out: bool
        True if ht_dtype is a complex float, False otherwise
    """
    return ht_dtype in _complexfloating


def heat_type_of(
    obj: Union[str, Type[datatype], Any, Iterable[str, Type[datatype], Any]]
) -> Type[datatype]:
    """
    Returns the corresponding HeAT data type of given object, i.e. scalar, array or iterable. Attempts to determine the
    canonical data type based on the following priority list:
        1. dtype property
        2. type(obj)
        3. type(obj[0])

    Parameters
    ----------
    obj : scalar or DNDarray or iterable
        The object for which to infer the type.

    Raises
    -------
    TypeError
        If the object's type cannot be inferred.
    """
    # attempt to access the dtype property
    try:
        return canonical_heat_type(obj.dtype)
    except (AttributeError, TypeError):
        pass

    # attempt type of object itself
    try:
        return canonical_heat_type(type(obj))
    except TypeError:
        pass

    # last resort, type of the object at first position
    try:
        return canonical_heat_type(type(obj[0]))
    except (KeyError, IndexError, TypeError):
        raise TypeError(f"data type of {obj} is not understood")


# type code assignment
__type_codes = collections.OrderedDict(
    [
        (bool, 0),
        (uint8, 1),
        (int8, 2),
        (int16, 3),
        (int32, 4),
        (int64, 5),
        (float32, 6),
        (float64, 7),
        (complex64, 8),
        (complex128, 9),
    ]
)

# safe cast table
__safe_cast = [
    # bool  uint8  int8   int16  int32  int64  float32 float64 complex64 complex128
    [True, True, True, True, True, True, True, True, True, True],  # bool
    [False, True, False, True, True, True, True, True, True, True],  # uint8
    [False, False, True, True, True, True, True, True, True, True],  # int8
    [False, False, False, True, True, True, True, True, True, True],  # int16
    [False, False, False, False, True, True, False, True, False, True],  # int32
    [False, False, False, False, False, True, False, True, False, True],  # int64
    [False, False, False, False, False, False, True, True, True, True],  # float32
    [False, False, False, False, False, False, False, True, False, True],  # float64
    [False, False, False, False, False, False, False, False, True, True],  # complex64
    [False, False, False, False, False, False, False, False, False, True],  # complex128
]

# intuitive cast table
__intuitive_cast = [
    # bool  uint8  int8   int16  int32  int64  float32 float64 complex64 complex128
    [True, True, True, True, True, True, True, True, True, True],  # bool
    [False, True, False, True, True, True, True, True, True, True],  # uint8
    [False, False, True, True, True, True, True, True, True, True],  # int8
    [False, False, False, True, True, True, True, True, True, True],  # int16
    [False, False, False, False, True, True, True, True, True, True],  # int32
    [False, False, False, False, False, True, False, True, False, True],  # int64
    [False, False, False, False, False, False, True, True, True, True],  # float32
    [False, False, False, False, False, False, False, True, False, True],  # float64
    [False, False, False, False, False, False, False, False, True, True],  # complex64
    [False, False, False, False, False, False, False, False, False, True],  # complex128
]


# same kind table
__same_kind = [
    # bool  uint8  int8   int16  int32  int64  float32 float64 complex64 complex128
    [True, False, False, False, False, False, False, False, False, False],  # bool
    [False, True, True, True, True, True, False, False, False, False],  # uint8
    [False, True, True, True, True, True, False, False, False, False],  # int8
    [False, True, True, True, True, True, False, False, False, False],  # int16
    [False, True, True, True, True, True, False, False, False, False],  # int32
    [False, True, True, True, True, True, False, False, False, False],  # int64
    [False, False, False, False, False, False, True, True, False, False],  # float32
    [False, False, False, False, False, False, True, True, False, False],  # float64
    [False, False, False, False, False, False, False, False, True, True],  # complex64
    [False, False, False, False, False, False, False, False, True, True],  # complex128
]


# static list of possible casting methods
__cast_kinds = ["no", "safe", "same_kind", "unsafe", "intuitive"]


def can_cast(
    from_: Union[str, Type[datatype], Any],
    to: Union[str, Type[datatype], Any],
    casting: str = "intuitive",
) -> bool:
    """
    Returns True if cast between data types can occur according to the casting rule. If from is a scalar or array
    scalar, also returns True if the scalar value can be cast without overflow or truncation to an integer.

    Parameters
    ----------
    from_ : Union[str, Type[datatype], Any]
        Scalar, data type or type specifier to cast from.
    to : Union[str, Type[datatype], Any]
        Target type to cast to.
    casting: str, optional
        options: {"no", "safe", "same_kind", "unsafe", "intuitive"}, optional
        Controls the way the cast is evaluated
            * "no" the types may not be cast, i.e. they need to be identical
            * "safe" allows only casts that can preserve values with complete precision
            * "same_kind" safe casts are possible and down_casts within the same type family, e.g. int32 -> int8
            * "unsafe" means any conversion can be performed, i.e. this casting is always possible
            * "intuitive" allows all of the casts of safe plus casting from int32 to float32


    Raises
    -------
    TypeError
        If the types are not understood or casting is not a string
    ValueError
        If the casting rule is not understood

    Examples
    --------
    >>> ht.can_cast(ht.int32, ht.int64)
    True
    >>> ht.can_cast(ht.int64, ht.float64)
    True
    >>> ht.can_cast(ht.int16, ht.int8)
    False
    >>> ht.can_cast(1, ht.float64)
    True
    >>> ht.can_cast(2.0e200, "u1")
    False
    >>> ht.can_cast('i8', 'i4', 'no')
    False
    >>> ht.can_cast("i8", "i4", "safe")
    False
    >>> ht.can_cast("i8", "i4", "same_kind")
    True
    >>> ht.can_cast("i8", "i4", "unsafe")
    True
    """
    if not isinstance(casting, str):
        raise TypeError(f"expected string, found {type(casting)}")
    if casting not in __cast_kinds:
        raise ValueError(f"casting must be one of {str(__cast_kinds)[1:-1]}")

    # obtain the types codes of the canonical HeAT types
    try:
        typecode_from = __type_codes[canonical_heat_type(from_)]
    except TypeError:
        typecode_from = __type_codes[heat_type_of(from_)]
    typecode_to = __type_codes[canonical_heat_type(to)]

    # unsafe casting allows everything
    if casting == "unsafe":
        return True

    # types have to match exactly
    elif casting == "no":
        return typecode_from == typecode_to

    # safe casting or same_kind
    can_safe_cast = __safe_cast[typecode_from][typecode_to]
    if casting == "safe":
        return can_safe_cast
    can_intuitive_cast = __intuitive_cast[typecode_from][typecode_to]
    if casting == "intuitive":
        return can_intuitive_cast
    return can_safe_cast or __same_kind[typecode_from][typecode_to]


# compute possible type promotions dynamically
__type_promotions = [[None] * len(row) for row in __same_kind]
for i, operand_a in enumerate(__type_codes.keys()):
    for j, operand_b in enumerate(__type_codes.keys()):
        for target in __type_codes.keys():
            if can_cast(operand_a, target) and can_cast(operand_b, target):
                __type_promotions[i][j] = target
                break


def iscomplex(x: dndarray.DNDarray) -> dndarray.DNDarray:
    """
    Test element-wise if input is complex.

    Parameters
    ----------
    x : DNDarray
        The input DNDarray

    Examples
    --------
    >>> ht.iscomplex(ht.array([1+1j, 1]))
    DNDarray([ True, False], dtype=ht.bool, device=cpu:0, split=None)
    """
    sanitation.sanitize_in(x)

    if issubclass(x.dtype, _complexfloating):
        return x.imag != 0
    else:
        return factories.zeros(x.shape, bool, split=x.split, device=x.device, comm=x.comm)


def isreal(x: dndarray.DNDarray) -> dndarray.DNDarray:
    """
    Test element-wise if input is real-valued.

    Parameters
    ----------
    x : DNDarray
        The input DNDarray

    Examples
    --------
    >>> ht.iscomplex(ht.array([1+1j, 1]))
    DNDarray([ True, False], dtype=ht.bool, device=cpu:0, split=None)
    """
    return _operations.__local_op(torch.isreal, x, None, no_cast=True)


def issubdtype(
    arg1: Union[str, Type[datatype], Any], arg2: Union[str, Type[datatype], Any]
) -> builtins.bool:
    """
    Returns True if first argument is a typecode lower/equal in type hierarchy.

    Parameters
    ----------
    arg1 : type, str, ht.dtype
        A description representing the type. It may be a a Python builtin type, string or an HeAT type already.
    arg2 : type, str, ht.dtype
        A description representing the type. It may be a a Python builtin type, string or an HeAT type already.


    Examples
    --------
    >>> ints = ht.array([1, 2, 3], dtype=ht.int32)
    >>> ht.issubdtype(ints.dtype, ht.integer)
    True
    >>> ht.issubdype(ints.dtype, ht.floating)
    False
    >>> ht.issubdtype(ht.float64, ht.float32)
    False
    >>> ht.issubdtype('i', ht.integer)
    True
    """
    # Assure that each argument is a ht.dtype
    arg1 = canonical_heat_type(arg1)
    arg2 = canonical_heat_type(arg2)

    return issubclass(arg1, arg2)


def promote_types(
    type1: Union[str, Type[datatype], Any], type2: Union[str, Type[datatype], Any]
) -> Type[datatype]:
    """
    Returns the data type with the smallest size and smallest scalar kind to which both ``type1`` and ``type2`` may be
    intuitively cast to, where intuitive casting refers to maintaining the same bit length if possible. This function
    is symmetric.

    Parameters
    ----------
    type1 : type or str or datatype
        type of first operand
    type2 : type or str or datatype
        type of second operand

    Examples
    --------
    >>> ht.promote_types(ht.uint8, ht.uint8)
    <class 'heat.core.types.uint8'>
    >>> ht.promote_types(ht.int32, ht.float32)
    <class 'heat.core.types.float32'>
    >>> ht.promote_types(ht.int8, ht.uint8)
    <class 'heat.core.types.int16'>
    >>> ht.promote_types("i8", "f4")
    <class 'heat.core.types.float64'>
    """
    typecode_type1 = __type_codes[canonical_heat_type(type1)]
    typecode_type2 = __type_codes[canonical_heat_type(type2)]

    return __type_promotions[typecode_type1][typecode_type2]


def result_type(
    *arrays_and_types: Tuple[Union[dndarray.DNDarray, Type[datatype], Any]]
) -> Type[datatype]:
    """
    Returns the data type that results from type promotions rules performed in an arithmetic operation.

    Parameters
    ----------
    arrays_and_types: List of arrays and types
        Input arrays, types or numbers of the operation.

    Examples
    --------
    >>> ht.result_type(ht.array([1], dtype=ht.int32), 1)
    ht.int32
    >>> ht.result_type(ht.float32, ht.array(1, dtype=ht.int8))
    ht.float32
    >>> ht.result_type("i8", "f4")
    ht.float64
    """

    def result_type_rec(*arrays_and_types):
        # derive type and set precedence (lower number, higher precedence)
        arg = arrays_and_types[0]

        try:
            # array / tensor
            if isinstance(arg, np.ndarray):
                type1 = canonical_heat_type(arg.dtype.char)
            else:
                type1 = canonical_heat_type(arg.dtype)

            if len(arg.shape) > 0:
                prec1 = 0  # array
            else:
                prec1 = 2  # scalar
        except (AttributeError, TypeError):
            try:
                # type
                if isinstance(arg, np.dtype):
                    arg = arg.char
                type1 = canonical_heat_type(arg)
                prec1 = 1
            except TypeError:
                # type instance
                type1 = canonical_heat_type(type(arg))
                prec1 = 3

        # multiple arguments
        if len(arrays_and_types) > 1:
            type2, prec2 = result_type_rec(*arrays_and_types[1:])

            # fast check same type
            if type1 == type2:
                return type1, min(prec1, prec2)
            # fast check same precedence
            if prec1 == prec2:
                return promote_types(type1, type2), prec1

            # check if parent type is identical and decide by precedence
            for sclass in (bool, integer, floating, complex):
                if issubdtype(type1, sclass) and issubdtype(type2, sclass):
                    if prec1 < prec2:
                        return type1, min(prec1, prec2)
                    else:
                        return type2, min(prec1, prec2)

            # different parent type: bool < int < float < complex
            tc1 = __type_codes[type1]
            tc2 = __type_codes[type2]

            if tc1 < tc2:
                return type2, min(prec1, prec2)
            else:
                return type1, min(prec1, prec2)

        # single argument
        return type1, prec1

    return result_type_rec(*arrays_and_types)[0]


class finfo:
    """
    Class describing machine limits (bit representation) of floating point types.

    Attributes
    ----------
    bits : int
        The number of bits occupied by the type.
    eps : float
        The smallest representable positive number such that
        ``1.0 + eps != 1.0``.  Type of ``eps`` is an appropriate floating
        point type.
    max : float
        The largest representable number.
    min : float
        The smallest representable number, typically ``-max``.
    tiny : float
        The smallest positive usable number.  Type of ``tiny`` is an
        appropriate floating point type.

    Parameters
    ----------
    dtype : datatype
        Kind of floating point data-type about which to get information.

    Examples
    ---------
    >>> import heat as ht
    >>> info = ht.types.finfo(ht.float32)
    >>> info.bits
    32
    >>> info.eps
    1.1920928955078125e-07
    """

    def __new__(cls, dtype: Type[datatype]):
        try:
            dtype = heat_type_of(dtype)
        except (KeyError, IndexError, TypeError):
            # If given type is not heat type
            pass

        if dtype not in _inexact:
            raise TypeError(f"Data type {dtype} not inexact, not supported")

        return super(finfo, cls).__new__(cls)._init(dtype)

    def _init(self, dtype: Type[datatype]):
        _torch_finfo = torch.finfo(dtype.torch_type())
        for word in ["bits", "eps", "max", "min", "tiny"]:
            setattr(self, word, getattr(_torch_finfo, word))

        return self


class iinfo:
    """
    Class describing machine limits (bit representation) of integer types.

    Attributes
    ----------
    bits : int
        The number of bits occupied by the type.
    max : float
        The largest representable number.
    min : float
        The smallest representable number, typically ``-max``.

    Parameters
    ----------
    dtype : datatype
        Kind of floating point data-type about which to get information.

    Examples
    ---------
    >>> import heat as ht
    >>> info = ht.types.iinfo(ht.int32)
    >>> info.bits
    32
    """

    def __new__(cls, dtype: Type[datatype]):
        try:
            dtype = heat_type_of(dtype)
        except (KeyError, IndexError, TypeError):
            # If given type is not heat type
            pass

        if dtype not in _exact:
            raise TypeError(f"Data type {dtype} not exact, not supported")

        return super(iinfo, cls).__new__(cls)._init(dtype)

    def _init(self, dtype: Type[datatype]):
        _torch_iinfo = torch.iinfo(dtype.torch_type())
        for word in ["bits", "min", "max"]:
            setattr(self, word, getattr(_torch_iinfo, word))

        return self


# dndarray is imported at the very end to break circular dependency
from . import dndarray



"""
Module Implementing the Kmeans Algorithm
"""
from typing import Optional, Union, TypeVar

import heat as ht
from heat.cluster._kcluster import _KCluster
from heat.core.dndarray import DNDarray

self = TypeVar("self")


class KMeans(_KCluster):
    r"""
    K-Means clustering algorithm. An implementation of Lloyd's algorithm [1].

    Attributes
    ----------
    n_clusters : int
        The number of clusters to form as well as the number of centroids to generate.
    init : str or DNDarray
        Method for initialization:

        - ‘k-means++’ : selects initial cluster centers for the clustering in a smart way to speed up convergence [2].
        - ‘random’: choose k observations (rows) at random from data for the initial centroids.
        - DNDarray: it should be of shape (n_clusters, n_features) and gives the initial centers.
    max_iter : int
        Maximum number of iterations of the k-means algorithm for a single run.
    tol : float
        Relative tolerance with regards to inertia to declare convergence.
    random_state : int
        Determines random number generation for centroid initialization.

    Notes
    -----
    The average complexity is given by :math:`O(k \cdot n \cdot T)`, were n is the number of samples and :math:`T` is the number of iterations.
    In practice, the k-means algorithm is very fast, but it may fall into local minima. That is why it can be useful
    to restart it several times. If the algorithm stops before fully converging (because of ``tol`` or ``max_iter``),
    ``labels_`` and ``cluster_centers_`` will not be consistent, i.e. the ``cluster_centers_`` will not be the means of the
    points in each cluster. Also, the estimator will reassign ``labels_`` after the last iteration to make ``labels_``
    consistent with predict on the training set.

    References
    ----------
    [1] Lloyd, Stuart P., "Least squares quantization in PCM", IEEE Transactions on Information Theory, 28 (2), pp.
    129–137, 1982.

    [2] Arthur, D., Vassilvitskii, S., "k-means++: The Advantages of Careful Seeding", Proceedings of the Eighteenth
    Annual ACM-SIAM Symposium on Discrete Algorithms, Society for Industrial and Applied Mathematics
    Philadelphia, PA, USA. pp. 1027–1035, 2007.
    """

    def __init__(
        self,
        n_clusters: int = 8,
        init: Union[str, DNDarray] = "random",
        max_iter: int = 300,
        tol: float = 1e-4,
        random_state: Optional[int] = None,
    ):
        if init == "kmeans++":
            init = "probability_based"

        super().__init__(
            metric=lambda x, y: ht.spatial.distance.cdist(x, y, quadratic_expansion=True),
            n_clusters=n_clusters,
            init=init,
            max_iter=max_iter,
            tol=tol,
            random_state=random_state,
        )

    def _update_centroids(self, x: DNDarray, matching_centroids: DNDarray):
        """
        Compute coordinates of new centroid as mean of the data points in ``x`` that are assigned to this centroid.

        Parameters
        ----------
        x :  DNDarray
            Input data
        matching_centroids : DNDarray
            Array filled with indices ``i`` indicating to which cluster ``ci`` each sample point in ``x`` is assigned

        """
        new_cluster_centers = self._cluster_centers.copy()
        for i in range(self.n_clusters):
            # points in current cluster
            selection = (matching_centroids == i).astype(ht.int64)
            # accumulate points and total number of points in cluster
            assigned_points = x * selection
            points_in_cluster = selection.sum(axis=0, keepdims=True).clip(
                1.0, ht.iinfo(ht.int64).max
            )

            # compute the new centroids
            new_cluster_centers[i : i + 1, :] = (assigned_points / points_in_cluster).sum(
                axis=0, keepdims=True
            )

        return new_cluster_centers

    def fit(self, x: DNDarray) -> self:
        """
        Computes the centroid of a k-means clustering.

        Parameters
        ----------
        x : DNDarray
            Training instances to cluster. Shape = (n_samples, n_features)

        """
        # input sanitation
        if not isinstance(x, DNDarray):
            raise ValueError(f"input needs to be a ht.DNDarray, but was {type(x)}")

        # initialize the clustering
        self._initialize_cluster_centers(x)
        self._n_iter = 0
        matching_centroids = ht.zeros((x.shape[0]), split=x.split, device=x.device, comm=x.comm)

        # iteratively fit the points to the centroids
        for epoch in range(self.max_iter):
            # increment the iteration count
            self._n_iter += 1
            # determine the centroids

            matching_centroids = self._assign_to_cluster(x)

            # update the centroids
            new_cluster_centers = self._update_centroids(x, matching_centroids)
            # check whether centroid movement has converged
            self._inertia = ((self._cluster_centers - new_cluster_centers) ** 2).sum()
            self._cluster_centers = new_cluster_centers.copy()
            if self.tol is not None and self._inertia <= self.tol:
                break

        self._labels = matching_centroids

        return self



"""
Learning rate schedulers in the heat namespace
"""

import sys
import torch.optim.lr_scheduler as lrs


def __getattr__(name):
    """
    Call the torch learning rate scheduler of a specified name
    """
    try:
        return lrs.__getattribute__(name)
    except AttributeError:
        raise AttributeError(f"name {name} is not implemented in torch.optim.lr_scheduler")



"""Manipulation operations for (potentially distributed) `DCSR_matrix`."""
from __future__ import annotations

from heat.sparse.dcsr_matrix import DCSR_matrix

from ..core.memory import sanitize_memory_layout
from ..core.dndarray import DNDarray
from ..core.factories import empty

__all__ = [
    "todense",
]


def todense(sparse_matrix: DCSR_matrix, order="C", out: DNDarray = None) -> DNDarray:
    """
    Convert :class:`~heat.sparse.DCSR_matrix` to a dense :class:`~heat.core.DNDarray`.
    Output follows the same distribution among processes as the input

    Parameters
    ----------
    sparse_matrix : :class:`~heat.sparse.DCSR_matrix`
        The sparse csr matrix which is to be converted to a dense array
    order: str, optional
        Options: ``'C'`` or ``'F'``. Specifies the memory layout of the newly created `DNDarray`. Default is ``order='C'``,
        meaning the array will be stored in row-major order (C-like). If ``order=‘F’``, the array will be stored in
        column-major order (Fortran-like).
    out : DNDarray
        Output buffer in which the values of the dense format is stored.
        If not specified, a new DNDarray is created.

    Raises
    ------
    ValueError
        If shape of output buffer does not match that of the input.
    ValueError
        If split axis of output buffer does not match that of the input.

    Examples
    --------
    >>> indptr = torch.tensor([0, 2, 3, 6])
    >>> indices = torch.tensor([0, 2, 2, 0, 1, 2])
    >>> data = torch.tensor([1, 2, 3, 4, 5, 6], dtype=torch.float)
    >>> torch_sparse_csr = torch.sparse_csr_tensor(indptr, indices, data)
    >>> heat_sparse_csr = ht.sparse.sparse_csr_matrix(torch_sparse_csr, split=0)
    >>> heat_sparse_csr
    (indptr: tensor([0, 2, 3, 6]), indices: tensor([0, 2, 2, 0, 1, 2]), data: tensor([1., 2., 3., 4., 5., 6.]), dtype=ht.float32, device=cpu:0, split=0)
    >>> heat_sparse_csr.todense()
    DNDarray([[1., 0., 2.],
              [0., 0., 3.],
              [4., 5., 6.]], dtype=ht.float32, device=cpu:0, split=0)
    """
    if out is not None:
        if out.shape != sparse_matrix.shape:
            raise ValueError(
                f"Expected output buffer with shape {sparse_matrix.shape} but was {out.shape}"
            )

        if out.split != sparse_matrix.split:
            raise ValueError(
                f"Expected output buffer with split axis {sparse_matrix.split} but was {out.split}"
            )

    if out is None:
        out = empty(
            shape=sparse_matrix.shape,
            split=sparse_matrix.split,
            dtype=sparse_matrix.dtype,
            device=sparse_matrix.device,
            comm=sparse_matrix.comm,
            order=order,
        )

    out.larray = sanitize_memory_layout(sparse_matrix.larray.to_dense(), order=order)
    return out


DCSR_matrix.todense = lambda self, order="C", out=None: todense(self, order, out)
DCSR_matrix.to_dense = lambda self, order="C", out=None: todense(self, order, out)



"""
This is the heat.nn submodule.

It contains data parallel specific nn modules. It also includes all of the modules in the torch.nn namespace
"""

import sys
import torch
import unittest

from . import functional


if sys.version_info.minor >= 7:
    from .data_parallel import *

    functional.__getattr__ = functional.func_getattr

    def __getattr__(name):
        """
        When a function is called for the heat.nn module it will attempt to run the heat nn module with that
        name, then, if there is no such heat nn module, it will attempt to get the torch nn module of that name.
        """
        torch_all = torch.nn.modules.__all__
        if name in torch_all:
            return torch.nn.__getattribute__(name)
        else:
            try:
                unittest.__getattribute__(name)
            except AttributeError:
                raise AttributeError(f"module {name} not implemented in Torch or Heat")

else:
    from . import data_parallel
    from . import tests

    class Wrapper(object):
        """
        Wrapper to handle the dynamic calling of torch.nn modules in the heat namespace
        """

        def __init__(self, wrapped):  # noqa: D107
            self.wrapped = wrapped
            self.torch_all = torch.nn.modules.__all__
            self.data_parallel_all = data_parallel.__all__

        def __getattr__(self, name):
            """
            When a function is called for the heat.nn module it will attempt to run the heat nn module with that
            name, then, if there is no such heat nn module, it will attempt to get the torch nn module of that name.
            """
            if name in self.torch_all:
                return torch.nn.__getattribute__(name)
            elif name == "functional":
                return functional
            elif name in self.data_parallel_all:
                return data_parallel.__getattribute__(name)
            elif name == "tests":
                return tests
            else:
                try:
                    unittest.__getattribute__(name)
                except AttributeError:
                    raise AttributeError(f"module '{name}' not implemented in Torch or Heat")

    sys.modules[__name__] = Wrapper(sys.modules[__name__])



"""Allows to output DNDarrays to stdout."""

import builtins
import copy
import torch
from .communication import MPI_WORLD

from .dndarray import DNDarray

__all__ = ["get_printoptions", "global_printing", "local_printing", "print0", "set_printoptions"]


# set the default printing width to a 120
_DEFAULT_LINEWIDTH = 120
torch.set_printoptions(profile="default", linewidth=_DEFAULT_LINEWIDTH)
LOCAL_PRINT = False

# printing
__PREFIX = "DNDarray"
__INDENT = len(__PREFIX)


def get_printoptions() -> dict:
    """
    Returns the currently configured printing options as key-value pairs.
    """
    return copy.copy(torch._tensor_str.PRINT_OPTS.__dict__)


def local_printing() -> None:
    """
    The builtin `print` function will now print the local PyTorch Tensor values for
    `DNDarrays` given as arguments.

    Examples
    --------
    >>> x = ht.ht.arange(15 * 5, dtype=ht.float).reshape((15, 5)).resplit(0)
    >>> ht.local_printing()
    [0/2]Printing options set to LOCAL. DNDarrays will print the local PyTorch Tensors
    >>> print(x)
    [0/2] [[ 0.,  1.,  2.,  3.,  4.],
    [0/2]  [ 5.,  6.,  7.,  8.,  9.],
    [0/2]  [10., 11., 12., 13., 14.],
    [0/2]  [15., 16., 17., 18., 19.],
    [0/2]  [20., 21., 22., 23., 24.]]
    [1/2] [[25., 26., 27., 28., 29.],
    [1/2]  [30., 31., 32., 33., 34.],
    [1/2]  [35., 36., 37., 38., 39.],
    [1/2]  [40., 41., 42., 43., 44.],
    [1/2]  [45., 46., 47., 48., 49.]]
    [2/2] [[50., 51., 52., 53., 54.],
    [2/2]  [55., 56., 57., 58., 59.],
    [2/2]  [60., 61., 62., 63., 64.],
    [2/2]  [65., 66., 67., 68., 69.],
    [2/2]  [70., 71., 72., 73., 74.]]
    """
    global LOCAL_PRINT
    LOCAL_PRINT = True
    print0("Printing options set to LOCAL. DNDarrays will print the local PyTorch Tensors")


def global_printing() -> None:
    """
    For `DNDarray`s, the builtin `print` function will gather all of the data, format it
    then print it on ONLY rank 0.

    Returns
    -------
    None

    Examples
    --------
    >>> x = ht.arange(15 * 5, dtype=ht.float).reshape((15, 5)).resplit(0)
    >>> print(x)
    [0] DNDarray([[ 0.,  1.,  2.,  3.,  4.],
                 [ 5.,  6.,  7.,  8.,  9.],
                 [10., 11., 12., 13., 14.],
                 [15., 16., 17., 18., 19.],
                 [20., 21., 22., 23., 24.],
                 [25., 26., 27., 28., 29.],
                 [30., 31., 32., 33., 34.],
                 [35., 36., 37., 38., 39.],
                 [40., 41., 42., 43., 44.],
                 [45., 46., 47., 48., 49.],
                 [50., 51., 52., 53., 54.],
                 [55., 56., 57., 58., 59.],
                 [60., 61., 62., 63., 64.],
                 [65., 66., 67., 68., 69.],
                 [70., 71., 72., 73., 74.]], dtype=ht.float32, device=cpu:0, split=0)
    """
    global LOCAL_PRINT
    if not LOCAL_PRINT:
        return
    LOCAL_PRINT = False
    print0(
        "Printing options set to GLOBAL. DNDarrays will be collected on process 0 before printing"
    )


def print0(*args, **kwargs) -> None:
    """
    Wraps the builtin `print` function in such a way that it will only run the command on
    rank 0. If this is called with DNDarrays and local printing, only the data local to
    process 0 is printed. For more information see the examples.

    This function is also available as a builtin when importing heat.

    Examples
    --------
    >>> x = ht.arange(15 * 5, dtype=ht.float).reshape((15, 5)).resplit(0)
    >>> # GLOBAL PRINTING
    >>> ht.print0(x)
    [0] DNDarray([[ 0.,  1.,  2.,  3.,  4.],
                 [ 5.,  6.,  7.,  8.,  9.],
                 [10., 11., 12., 13., 14.],
                 [15., 16., 17., 18., 19.],
                 [20., 21., 22., 23., 24.],
                 [25., 26., 27., 28., 29.],
                 [30., 31., 32., 33., 34.],
                 [35., 36., 37., 38., 39.],
                 [40., 41., 42., 43., 44.],
                 [45., 46., 47., 48., 49.],
                 [50., 51., 52., 53., 54.],
                 [55., 56., 57., 58., 59.],
                 [60., 61., 62., 63., 64.],
                 [65., 66., 67., 68., 69.],
                 [70., 71., 72., 73., 74.]], dtype=ht.float32, device=cpu:0, split=0)
    >>> ht.local_printing()
    [0/2] Printing options set to LOCAL. DNDarrays will print the local PyTorch Tensors
    >>> print0(x)
    [0/2] [[ 0.,  1.,  2.,  3.,  4.],
    [0/2]  [ 5.,  6.,  7.,  8.,  9.],
    [0/2]  [10., 11., 12., 13., 14.],
    [0/2]  [15., 16., 17., 18., 19.],
    [0/2]  [20., 21., 22., 23., 24.]], device: cpu:0, split: 0
    """
    if not LOCAL_PRINT:
        args = list(args)
        for i in range(len(args)):
            if isinstance(args[i], DNDarray):
                args[i] = __str__(args[i])
        args = tuple(args)
    if MPI_WORLD.rank == 0:
        print(*args, **kwargs)


builtins.print0 = print0


def set_printoptions(
    precision=None, threshold=None, edgeitems=None, linewidth=None, profile=None, sci_mode=None
):
    """
    Configures the printing options. List of items shamelessly taken from NumPy and PyTorch (thanks guys!).

    Parameters
    ----------
    precision : int, optional
        Number of digits of precision for floating point output (default=4).
    threshold : int, optional
        Total number of array elements which trigger summarization rather than full `repr` string (default=1000).
    edgeitems : int, optional
        Number of array items in summary at beginning and end of each dimension (default=3).
    linewidth : int, optional
        The number of characters per line for the purpose of inserting line breaks (default = 80).
    profile : str, optional
        Sane defaults for pretty printing. Can override with any of the above options. Can be any one of `default`,
        `short`, `full`.
    sci_mode : bool, optional
        Enable (True) or disable (False) scientific notation. If None (default) is specified, the value is automatically
        inferred by HeAT.
    """
    torch.set_printoptions(precision, threshold, edgeitems, linewidth, profile, sci_mode)

    # HeAT profiles will print a bit wider than PyTorch does
    if (
        (profile == "default" and linewidth is None)
        or (profile == "short" and linewidth is None)
        or (profile == "full" and linewidth is None)
    ):
        torch._tensor_str.PRINT_OPTS.linewidth = _DEFAULT_LINEWIDTH


def __str__(dndarray) -> str:
    """
    Computes a printable representation of the passed DNDarray.

    Parameters
    ----------
    dndarray : DNDarray
        The array for which to obtain the corresponding string
    """
    if LOCAL_PRINT:
        return (
            torch._tensor_str._tensor_str(dndarray.larray, 0)
            + f", device: {dndarray.device}, split: {dndarray.split}"
        )
    tensor_string = _tensor_str(dndarray, __INDENT + 1)
    if dndarray.comm.rank != 0:
        return ""

    return (
        f"{__PREFIX}({tensor_string}, dtype=ht.{dndarray.dtype.__name__}, "
        f"device={dndarray.device}, split={dndarray.split})"
    )


def _torch_data(dndarray, summarize) -> DNDarray:
    """
    Extracts the data to be printed from the DNDarray in form of a torch tensor and returns it.

    Parameters
    ----------
    dndarray : DNDarray
        The HeAT DNDarray to be printed.
    summarize : bool
        Flag indicating whether to print the full data or summarized, i.e. ellipsed, version of the data.
    """
    if not dndarray.is_balanced():
        dndarray.balance_()
    # data is not split, we can use it as is
    if dndarray.split is None or dndarray.comm.size == 1:
        data = dndarray.larray
    # split, but no summary required, we collect it
    elif not summarize:
        data = dndarray.copy().resplit_(None).larray
    # split, but summarized, collect the slices from all nodes and pass it on
    else:
        edgeitems = torch._tensor_str.PRINT_OPTS.edgeitems
        double_items = 2 * edgeitems
        ndims = dndarray.ndim
        data = dndarray.larray

        for i in range(ndims):
            # skip over dimensions that are smaller than twice the number of edge items to display
            if dndarray.gshape[i] <= double_items:
                continue

            # non-split dimension, can slice locally
            if i != dndarray.split:
                start_tensor = torch.index_select(data, i, torch.arange(edgeitems + 1))
                end_tensor = torch.index_select(
                    data, i, torch.arange(dndarray.lshape[i] - edgeitems, dndarray.lshape[i])
                )
                data = torch.cat([start_tensor, end_tensor], dim=i)
            # split-dimension , need to respect the global offset
            elif i == dndarray.split and dndarray.gshape[i] > double_items:
                offset, _, _ = dndarray.comm.chunk(dndarray.gshape, i)

                if offset < edgeitems + 1:
                    end = min(dndarray.lshape[i], edgeitems + 1 - offset)
                    data = torch.index_select(data, i, torch.arange(end))
                elif dndarray.gshape[i] - edgeitems < offset - dndarray.lshape[i]:
                    global_start = dndarray.gshape[i] - edgeitems
                    data = torch.index_select(
                        data, i, torch.arange(max(0, global_start - offset), dndarray.lshape[i])
                    )
        # exchange data
        received = dndarray.comm.gather(data)
        if dndarray.comm.rank == 0:
            # concatenate data along the split axis
            data = torch.cat(received, dim=dndarray.split)

    return data


def _tensor_str(dndarray, indent: int) -> str:
    """
    Computes a string representation of the passed DNDarray.

    Parameters
    ----------
    dndarray: DNDarray
        The array for which to obtain the corresponding string
    indent: int
        The number of spaces the array content is indented.
    """
    elements = dndarray.gnumel
    if elements == 0:
        return "[]"

    # we will recycle torch's printing features here
    # to do so, we slice up the torch data and forward it to torch internal printing mechanism
    summarize = elements > get_printoptions()["threshold"]
    torch_data = _torch_data(dndarray, summarize)
    formatter = torch._tensor_str._Formatter(torch_data)

    return torch._tensor_str._tensor_str_with_formatter(torch_data, indent, summarize, formatter)
